/*
    Fichier majcycle.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Bibliothque de fonctions permettant de prendre en compte
    un changement de cycle ou de se positionner dans le cycle
    en cours pour un mise  jour du fichier cyloop.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "types.h"
#include "majcycle.h"


// variables globales (pour viter des tas de passages de paramtre)
extern FILE        *descfic;  // descripteur du fichier cyloop
extern desc_cyloop entetefic; // entte du fichier cyloop
extern desc_var    entetevar; // entte de la variable du fichier cyloop



// rcupration du nom du fichier cyloop et rajout ventuel de son suffixe

char *recup_nomfic (char *nom)
{
    char  *nomfic;
    int   lg_nomfic;


    // si un nom de fichier a t pass en paramtre
    if (nom)
    {
        // rcuprer sa taille
        lg_nomfic = strlen (nom);

        // si ce nom se termine par le suffixe .cyl
        if ((lg_nomfic > 4) && (strcmp (nom + lg_nomfic - 4, ".cyl") == 0))
        {
            // mmoriser le nom tel quel
            nomfic = malloc (lg_nomfic + 1);
            strcpy (nomfic, nom);
        }
        // sinon
        else
        {
            // rajouter au nom le suffixe .cyl et mmoriser le rsultat
            nomfic = malloc (lg_nomfic + 5);
            strcpy (nomfic, nom);
            strcpy (nomfic + lg_nomfic, ".cyl");
        }
    }

    // retourner le nom du fichier
    return nomfic;
}



/*
    Prend en compte un passage  l'heure d'hiver ou d't

    retourne le valeur OUI ou NON selon qu'il faille reporter
    une mise  jour d'entetefic dans le fichier cyloop
*/

int maj_type_heure (int passage_ete, long timep)
{
    // si on vient juste de passer  l'heure d't
    if (passage_ete)
    {
        // mmoriser le changement d'heure
        entetefic.etat_cyloop = entetefic.etat_cyloop | heure_ete;

        // et modifier l'instant de dbut du cycle
        entetefic.deb_cycle = entetefic.deb_cycle - 3600;

        // on reportera plus tard la mise  jour dans le fichier cyloop
        return (OUI);
    }
    // sinon on vient juste de passer  l'heure d'hiver
    else
    {
        // si ce changement d'heure ne fait pas revenir au cycle prcdent
        if (entetefic.deb_cycle + 3600 <= timep)
        {
            // mmoriser le changement d'heure
            entetefic.etat_cyloop = entetefic.etat_cyloop & mask_heure_hiver;

            // et modifier l'instant de dbut du cycle
            entetefic.deb_cycle = entetefic.deb_cycle + 3600;

            // on reportera plus tard la mise  jour dans le fichier cyloop
            return (OUI);
        }
        // sinon
        else
            // le changement d'heure sera pris en compte une autre fois
            return (NON);
    }
}



// prise en compte d'un changement de cycle

void changement_cycle (long timep)
{
    // structure contenant la date et l'heure du dbut du cycle
    struct tm oldstm;
    int    cycles_sautes; // compteur de cycles
    uint32 duree_cycle;   // dure du cycle
    int    coef_pond;     // coefficient de pondration  appliquer


    // initialisation, dure du cycle
    duree_cycle = entetefic.duree_cycle;

    // si cycle de longueur irrgulire bas sur le calendrier
    if (entetefic.caract_gen & cycle_ireg)
    {
        // dcouper les caractristique de l'instant du dbut de cycle
        if (entetefic.caract_gen & temps_utc)
            gmtime_r (&entetefic.deb_cycle, &oldstm);
        else
            localtime_r (&entetefic.deb_cycle, &oldstm);

        // initialisation compteur
        cycles_sautes = 0;

        // si cycle annuel
        if (entetefic.caract_gen & cycle_annuel)
        {
            // on avance anne par anne
            do
            {
                entetefic.deb_cycle += duree_cycle;

                if (++ oldstm.tm_year % 4)
                    duree_cycle = 365 * secjour;
                else
                    duree_cycle = 366 * secjour;

                cycles_sautes ++;
            }
            // jusqu' ce qu'on tombe sur la bonne
            while (entetefic.deb_cycle + duree_cycle <= timep);
        }
        // sinon, cycle mensuel
        else
        {
            // on avance mois par mois
            do
            {
                entetefic.deb_cycle += duree_cycle;

                switch (++ oldstm.tm_mon)
                {
                             // fvrier
                    case  1: if (oldstm.tm_year % 4)
                                 entetefic.duree_cycle = 28 * secjour;
                             else
                                 entetefic.duree_cycle = 29 * secjour;

                             break;

                    case  3: // mois de 30 jours
                    case  5:
                    case  8:
                    case 10: entetefic.duree_cycle = 30 * secjour;
                             break;

                             // janvier, changement d'anne
                    case 12: oldstm.tm_mon = 0;
                             oldstm.tm_year ++;

                             // mois de 31 jours (7 mois par an)
                    default: entetefic.duree_cycle = 31 * secjour;
                }

                cycles_sautes ++;
            }
            // jusqu' ce qu'on tombe sur le bon
            while (entetefic.deb_cycle + duree_cycle <= timep);
        }
    }
    // sinon (cas gnral, cycle de longueur rgulire)
    else
    {
        // calculer de combien de cycles on avance
        cycles_sautes = (timep - entetefic.deb_cycle) / duree_cycle;

        // calculer l'instant de dbut du cycle courant
        entetefic.deb_cycle += cycles_sautes * duree_cycle;
    }

    // si calcul de pondration
    if (entetevar.ponderation)
    {
        // calcul du coefficient de pondration
        coef_pond = entetevar.ponderation;

        if (entetevar.typedon & cpt_cycle_vide)
        {
            while (--cycles_sautes)
                coef_pond = (coef_pond * entetevar.ponderation) >> 8;
        }

        // appliquer le calcul de pondration sur la variable
        pondere (coef_pond);
    }
}



// Retourne la dure du cycle actuel en secondes
// saute si ncessaire le 29 fvrier dans les cycles annuels

uint32 duree_cycle_actuel (int32 *timep)
{
    struct tm stm;  // structure contenant la date et l'heure courante


    // si cycle calendaire de longueur irrgulire
    if (entetefic.caract_gen & cycle_ireg)
    {
        // dans le cas des cycles annuels
        if (entetefic.caract_gen & cycle_annuel)
        {
            // rcuprer la date courante
            if (entetefic.caract_gen & temps_utc)
                gmtime_r (timep, &stm);
            else
                localtime_r (timep, &stm);

            // sauter le 29 fvrier le cas chant
            if ((stm.tm_mon >= 2) && (stm.tm_year % 4))
                *timep = *timep + secjour;

            // l'anne est compt sur 366 jours pour le graphique
            return (366 * secjour);
        }
        // sinon, on est dans le cas des cycles mensuels
        else
            // le mois est compt sur 31 jours pour le graphique
            return (31 * secjour);
    }
    // sinon (cas gnral, cycle de longueur rgulire)
    else
        // on utilise la vraie dure du cycle
        return (entetefic.duree_cycle);
}



// calcul du numro de l'instant du cycle  mettre  jour

uint16 calcul_instant (uint32 timep, uint32 duree_cycle)
{
    uint32 depuis_deb_cycle; // temps coul depuis le dbut du cycle
    uint16 result;           // rsultat du calcul


    // temps coul depuis le dbut du cycle courant
    depuis_deb_cycle = timep - entetefic.deb_cycle;

    // si on peut calculer le numro d'instant
    // sans dbordement mathmatique
    if (((depuis_deb_cycle * entetefic.nb_don_cycle) / entetefic.nb_don_cycle)
                                                          == depuis_deb_cycle)
        // calcul exact en arithmtique entire
        result = depuis_deb_cycle * entetefic.nb_don_cycle / duree_cycle;
    else
        // sinon calcul approch en arithmtique flottante
        result = (float)depuis_deb_cycle * entetefic.nb_don_cycle / duree_cycle;

    // retour du rsultat, les numros d'instant
    // vont de 1  entetefic.nb_don_cycle
    return (result + 1);
}
